/* librawkey v0.1 - (c) 1994 Russell Marks
 * This library may be freely used/copied/modified provided this copyright
 * notice is left intact.
 *
 * needs keymap support in kernel - been there since 0.99pl12 I think.
 */

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/kd.h>		/* RAW mode stuff, etc. */
#include <linux/keyboard.h>	/* mainly for NR_KEYS */
#include <linux/vt.h>		/* for VT stuff - nah, really? :) */
#include "rawkey.h"


static int keymap[NR_KEYS];	/* scancode -> ASCII translation */
static int asciimap[256];	/* ASCII -> scancode translation */
static int tty_fd;

static need_waitactive=0;	/* by default, assume svgalib does this */

static int key_down[128];


/* it's really easy to translate the scancodes these days, we just
 * use the keytable stuff!
 *
 * returns 0 on error, 1 if ok
 */
int get_keyb_map()
{
struct kbentry keyb_ent;
int f;

keyb_ent.kb_table=0;	/* unshifted */
for(f=0;f<NR_KEYS;f++)
  {
  keyb_ent.kb_index=f;
  
  if(ioctl(tty_fd,KDGKBENT,(unsigned int)&keyb_ent))
    return(0);
    
  keymap[f]=keyb_ent.kb_value;
  }
return(1);
}


void make_ascii_map()
{
int f,i,found;

for(f=0;f<256;f++)
  {
  found=0;
  for(i=0;i<NR_KEYS;i++)
    if((KTYP(keymap[i])==KT_LATIN || KTYP(keymap[i])==KT_LETTER ||
        KTYP(keymap[i])==KT_ASCII) && KVAL(keymap[i])==f)
      {
      found=1;
      break;
      }
      
  asciimap[f]=found?i:-1;
  }
}


void raw_mode(tty_fd,on)
int tty_fd,on;
{
ioctl(tty_fd,KDSKBMODE,on?K_RAW:K_XLATE);
}


void blank_key_down()
{
int f;

for(f=0;f<NR_KEYS;f++)
  key_down[f]=0;
}


/* returns 1 if ok, 0 otherwise */
int rawmode_init()
{
if((tty_fd=open("/dev/tty0",O_RDONLY | O_NONBLOCK))<0)
  return(0);

if(get_keyb_map(tty_fd,keymap))
  {
  make_ascii_map();
  blank_key_down();
  raw_mode(tty_fd,1);
  return(1);
  }
else
  return(0);
}


void rawmode_exit()
{
raw_mode(tty_fd,0);
close(tty_fd);
}


/* returns -1 if no keypresses pending, else returns scancode. */
int get_scancode()
{
unsigned char c;

if(read(tty_fd,&c,1)<=0)
  return(-1);

return((int)c);
}


/* returns >0 if switched, 0 if couldn't because we're already *on* that one.
 * nicked from handlevt.c in zgv, but don't worry, I asked the author. :^)
 * we assume here that we're using svgalib which handles the WAITACTIVE and
 * screen saving stuff for us.
 */
int vt_switch(newvt)
int newvt;
{
int vtnow;
struct vt_stat vts;

ioctl(tty_fd,VT_GETSTATE,&vts);
vtnow=vts.v_active;
if(newvt!=vtnow)
  {
  ioctl(tty_fd,VT_ACTIVATE,newvt);
  return(vtnow);
  }
else
  return(0);
}


/* only call this after vt_switch() if you're not using svgalib, and
 * if vt_switch() returned >0. The arg is the return value from vt_switch().
 * it causes a sleep until our VT is active again.
 */
void vt_wait_active(old_vt)
int old_vt;
{
ioctl(tty_fd,VT_WAITACTIVE,old_vt);
}


int is_key_pressed(sc)
int sc;
{
if(sc<0 || sc>127) return(0);
return(key_down[sc]?1:0);
}


int is_any_key_pressed()
{
int f;

for(f=0;f<128;f++)
  if(key_down[f]) return(1);

return(0);
}


/* this is the routine you should call whenever you would normally
 * read a keypress. However, to actually tell if a key is pressed,
 * call is_key_pressed() with a scancode as arg.
 */
void scan_keyboard()
{
int c,key,flag,vt;

/* we use BFI to fix the PrtSc/Pause problem - i.e. we don't :^) */
while((c=get_scancode())==0xE0);
if(c==0xE1) c=get_scancode();

if(c==-1) return;	/* no key was pressed */

key=c&127;
flag=(c&128)?0:1;	/* 1 = down */

key_down[key]=flag;

#if 0		/* XXX why doesn't this work!? */
/* check for alt-Fn - ctrl+alt, shift+alt also count */
/* only F1 to F10 for now */
if(key_down[LEFT_ALT] && flag && key>=FUNC_KEY(1) && key<=FUNC_KEY(10))
  {
  /* XXX should we reset to XLATE mode !? */
  raw_mode(tty_fd,0);
  
  /* switch consoles */
  vt=vt_switch(key-FUNC_KEY(0));
    
  /* wait for switch back, if we need to */
  if(need_waitactive)
    vt_wait_active(vt);

  /* we don't know what to assume about the keyboard state, 
   * so assume the best :)
   */
  blank_key_down();
  raw_mode(tty_fd,1);
  }
#endif
}


/* converts scancode to key binding */
int keymap_trans(sc)
int sc;
{
if(sc<0 || sc>127) return(-1);
return(keymap[sc]);
}


/* converts ASCII to scancode */
int scancode_trans(asc)
int asc;
{
if(asc<0 || asc>255) return(-1);
return(asciimap[asc]);
}
